import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:osam/osam.dart';

typedef ValueMapper<ST, V> = V Function(ST state);
typedef StateMapper<ST> = void Function(ST state);

// ignore: must_be_immutable
abstract class BaseState<ST> extends Equatable {
  int _lastKnownHashCode = -1;

  // ignore: close_sinks
  final _stateBroadcaster = StreamController<ST>.broadcast();

  @override
  List<Object> get props;

  ValueStream<ST> get _stateStream => ValueStream(_stateBroadcaster.stream.map((state) => state), this as ST);

  ValueStream<V> propertyStream<V>(ValueMapper<ST, V> mapper) => ValueStream(
      (PropertyStream<V>(
              _stateStream.stream.map<V>((state) {
                try {
                  return mapper(state);
                } catch (e) {
                  handleError(e);
                  return null;
                }
              }).where((property) => property != null),
              mapper(_stateStream.value)))
          .propertyStream,
      mapper(_stateStream.value));

  // mutable props workaround for hive adaptor
  void refreshHashcode() => _lastKnownHashCode = hashCode;

  void update(StateMapper<ST> change) {
    try {
      change(this as ST);
    } catch (e) {
      handleError(e);
    }
    if (_lastKnownHashCode != hashCode) {
      _stateBroadcaster.sink.add(this as ST);
      _lastKnownHashCode = hashCode;
    }
  }

  Future<void> updateAsync(StateMapper<ST> change) async {
    try {
      change(this as ST);
    } catch (e) {
      handleError(e);
    }
    if (_lastKnownHashCode != hashCode) {
      final c = Completer<bool>();
      _stateBroadcaster.sink.add(this as ST);
      _lastKnownHashCode = hashCode;
      c.complete();
      await c.future;
    }
  }

  void handleError(Object error) => print(error);

  void destruct() => _stateBroadcaster.close();

  void throwError(Object error) => _stateBroadcaster.sink.addError(error);

  void clear() {}
}

class PropertyStream<V> {
  final Stream<V> _inputStream;
  final V _firstValue;
  int _valueHashCode = -1;

  PropertyStream(this._inputStream, this._firstValue) {
    _valueHashCode = _ComparableWrapper(_firstValue).hashCode;
  }

  Stream<V> get propertyStream => _inputStream.map((next) {
        final comparableWrapper = _ComparableWrapper(next);
        final result = _valueHashCode == comparableWrapper.hashCode;
        _valueHashCode = comparableWrapper.hashCode;
        return result ? null : next;
      }).where((event) => event != null);
}

class _ComparableWrapper<V> extends Equatable {
  final V _value;

  _ComparableWrapper(this._value);

  @override
  List<Object> get props => [_value];
}
